from reinvent_scoring.scoring.diversity_filters.curriculum_learning.diversity_filter_memory import DiversityFilterMemory
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.diversity_filter_parameters import \
    DiversityFilterParameters



from reinvent_scoring.scoring.diversity_filters.curriculum_learning import DiversityFilterParameters
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.base_diversity_filter import BaseDiversityFilter
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.identical_murcko_scaffold import \
    IdenticalMurckoScaffold
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.identical_topological_scaffold import \
    IdenticalTopologicalScaffold
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.no_filter import NoFilter
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.no_filter_with_penalty import NoFilterWithPenalty
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.scaffold_similarity import ScaffoldSimilarity


class DiversityFilter:

    def __new__(cls, parameters: DiversityFilterParameters) -> BaseDiversityFilter:
        all_filters = dict(IdenticalMurckoScaffold=IdenticalMurckoScaffold,
                           NoFilterWithPenalty=NoFilterWithPenalty,
                           IdenticalTopologicalScaffold=IdenticalTopologicalScaffold,
                           ScaffoldSimilarity=ScaffoldSimilarity,
                           NoFilter=NoFilter
                           )
        div_filter = all_filters.get(parameters.name, KeyError(f"Invalid filter name: `{parameters.name}'"))
        return div_filter(parameters)



from dataclasses import dataclass


@dataclass
class DiversityFilterParameters:
    name: str
    minscore: float = 0.4
    bucket_size: int = 25
    minsimilarity: float = 0.4
    penalty_multiplier: float = 0.5



from copy import deepcopy
import numpy as np

from reinvent_scoring.scoring.diversity_filters.curriculum_learning import DiversityFilterParameters
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.base_diversity_filter import BaseDiversityFilter
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.memory_record_dto import MemoryRecordDTO
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.update_diversity_filter_dto import \
    UpdateDiversityFilterDTO


class NoFilterWithPenalty(BaseDiversityFilter):
    """Penalize previously generated compounds."""

    def __init__(self, parameters: DiversityFilterParameters):
        super().__init__(parameters)

    def update_score(self, dto: UpdateDiversityFilterDTO) -> np.array:
        score_summary = deepcopy(dto.score_summary)
        scores = score_summary.total_score
        smiles = score_summary.scored_smiles

        for i in score_summary.valid_idxs:
            smiles[i] = self._chemistry.convert_to_rdkit_smiles(smiles[i])
            scores[i] = self.parameters.penalty_multiplier * scores[i] if self._smiles_exists(smiles[i]) else scores[i]

        for i in score_summary.valid_idxs:
            if scores[i] >= self.parameters.minscore:
                loggable_data = self._compose_loggable_data(dto.loggable_data[i])
                memory_dto = MemoryRecordDTO(i, dto.step, scores[i], smiles[i], smiles[i], loggable_data,
                                             score_summary.scaffold_log)
                self._add_to_memory(memory_dto)
        return scores



from typing import List, Dict

import pandas as pd

from reinvent_scoring.scoring.diversity_filters.curriculum_learning.column_names_enum import ColumnNamesEnum
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.memory_record_dto import MemoryRecordDTO
from reinvent_scoring.scoring.score_summary import ComponentSummary
from reinvent_scoring.scoring.enums.scoring_function_component_enum import ScoringFunctionComponentNameEnum


class DiversityFilterMemory:

    def __init__(self):
        self._sf_component_name = ScoringFunctionComponentNameEnum()
        self._column_name = ColumnNamesEnum()
        df_dict = {self._column_name.STEP: [], self._column_name.SCAFFOLD: [], self._column_name.SMILES: [],
                   self._column_name.METADATA: []}
        self._memory_dataframe = pd.DataFrame(df_dict)

    def update(self, dto: MemoryRecordDTO):
        component_scores = {c.parameters.name: float(c.total_score[dto.id]) for c in dto.components}
        component_scores = self._include_raw_score(dto.id, component_scores, dto.components)
        component_scores[self._sf_component_name.TOTAL_SCORE] = float(dto.score)
        if not self.smiles_exists(dto.smile): self._add_to_memory_dataframe(dto, component_scores)

    def _add_to_memory_dataframe(self, dto: MemoryRecordDTO, component_scores: Dict):
        data = []
        headers = []
        for name, score in component_scores.items():
            headers.append(name)
            data.append(score)
        headers.append(self._column_name.STEP)
        data.append(dto.step)
        headers.append(self._column_name.SCAFFOLD)
        data.append(dto.scaffold)
        headers.append(self._column_name.SMILES)
        data.append(dto.smile)
        headers.append(self._column_name.METADATA)
        data.append(dto.loggable_data)
        new_data = pd.DataFrame([data], columns=headers)
        self._memory_dataframe = pd.concat([self._memory_dataframe, new_data], ignore_index=True, sort=False)

    def get_memory(self) -> pd.DataFrame:
        return self._memory_dataframe

    def set_memory(self, memory: pd.DataFrame):
        self._memory_dataframe = memory

    def smiles_exists(self, smiles: str):
        if len(self._memory_dataframe) == 0:
            return False
        return smiles in self._memory_dataframe[self._column_name.SMILES].values

    def scaffold_instances_count(self, scaffold: str):
        return (self._memory_dataframe[self._column_name.SCAFFOLD].values == scaffold).sum()

    def number_of_scaffolds(self):
        return len(set(self._memory_dataframe[self._column_name.SCAFFOLD].values))

    def number_of_smiles(self):
        return len(set(self._memory_dataframe[self._column_name.SMILES].values))

    def _include_raw_score(self, indx: int, component_scores: dict, components: List[ComponentSummary]):
        raw_scores = {f'raw_{c.parameters.name}': float(c.raw_score[indx]) for c in components if
                      c.raw_score is not None}
        all_scores = {**component_scores, **raw_scores}
        return all_scores



from dataclasses import dataclass


@dataclass
class UpdateLoggableDataDTO:
    """This class is used by the Diversity Filters to log out metadata."""
    input: str
    output: str
    likelihood: float = None
    prior_likelihood: float = None



from dataclasses import dataclass
from typing import List

from reinvent_scoring.scoring.diversity_filters.curriculum_learning.loggable_data_dto import UpdateLoggableDataDTO
from reinvent_scoring.scoring.score_summary import FinalSummary


@dataclass
class UpdateDiversityFilterDTO:
    score_summary: FinalSummary
    loggable_data: List[UpdateLoggableDataDTO]
    step: int = 0



from copy import deepcopy

import numpy as np
from rdkit.Chem.Scaffolds import MurckoScaffold

from reinvent_scoring.scoring.diversity_filters.curriculum_learning import DiversityFilterParameters
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.base_diversity_filter import BaseDiversityFilter
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.memory_record_dto import MemoryRecordDTO
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.update_diversity_filter_dto import \
    UpdateDiversityFilterDTO


class IdenticalTopologicalScaffold(BaseDiversityFilter):
    """Penalizes compounds based on exact Topological Scaffolds previously generated."""

    def __init__(self, parameters: DiversityFilterParameters):
        super().__init__(parameters)

    def update_score(self, dto: UpdateDiversityFilterDTO) -> np.array:
        score_summary = deepcopy(dto.score_summary)
        scores = score_summary.total_score
        smiles = score_summary.scored_smiles

        for i in score_summary.valid_idxs:
            smile = self._chemistry.convert_to_rdkit_smiles(smiles[i])
            scaffold = self._calculate_scaffold(smile)
            scores[i] = 0 if self._smiles_exists(smile) else scores[i]

            if scores[i] >= self.parameters.minscore:
                loggable_data = self._compose_loggable_data(dto.loggable_data[i]) if dto.loggable_data else ''
                memory_dto = MemoryRecordDTO(i, dto.step, scores[i], smile, scaffold, loggable_data,
                                             score_summary.scaffold_log)
                self._add_to_memory(memory_dto)
                scores[i] = self._penalize_score(scaffold, scores[i])

        return scores

    def _calculate_scaffold(self, smile):
        mol = self._chemistry.smile_to_mol(smile)
        if mol:
            try:
                scaffold = MurckoScaffold.MakeScaffoldGeneric(MurckoScaffold.GetScaffoldForMol(mol))
                scaffold_smiles = self._chemistry.mol_to_smiles(scaffold)
            except ValueError:
                scaffold_smiles = ''
        else:
            scaffold_smiles = ''
        return scaffold_smiles



from copy import deepcopy

import numpy as np

from reinvent_scoring.scoring.diversity_filters.curriculum_learning import DiversityFilterParameters
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.base_diversity_filter import BaseDiversityFilter
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.memory_record_dto import MemoryRecordDTO
from reinvent_scoring.scoring.diversity_filters.curriculum_learning.update_diversity_filter_dto import \
    UpdateDiversityFilterDTO


class NoFilter(BaseDiversityFilter):
    """Doesn't penalize compounds."""

    def __init__(self, parameters: DiversityFilterParameters):
        super().__init__(parameters)

    def update_score(self, dto: UpdateDiversityFilterDTO) -> np.array:
        score_summary = deepcopy(dto.score_summary)
        scores = score_summary.total_score
        for i in score_summary.valid_idxs:
            if scores[i] >= self.parameters.minscore:
                #TODO: perhaps no validation is needed
                # smile = score_summary.scored_smiles[i]
                smile = self._chemistry.convert_to_rdkit_smiles(score_summary.scored_smiles[i])
                loggable_data = self._compose_loggable_data(dto.loggable_data[i]) if dto.loggable_data else ''
                memory_dto = MemoryRecordDTO(i, dto.step, scores[i], smile, smile, loggable_data,
                                             score_summary.scaffold_log)
                self._add_to_memory(memory_dto)
        return scores




from dataclasses import dataclass
from typing import List

from reinvent_scoring.scoring.score_summary import ComponentSummary


@dataclass
class MemoryRecordDTO:
    id: int
    step: int
    score: float
    smile: str
    scaffold: str
    loggable_data: str
    components: List[ComponentSummary]
